from dataloader import Dataset
import argparse
from torch_geometric.data import DataLoader
from sklearn.model_selection import train_test_split

from train import train
import torch
import random
import numpy as np
import time



TORCH_MAJOR = int(torch.__version__.split('.')[0])
TORCH_MINOR = int(torch.__version__.split('.')[1])
#t

parser = argparse.ArgumentParser()
parser.add_argument('--dataset', type=str, default='book-crossing', help='which dataset to use')
parser.add_argument('--rating_file', type=str, default='implicit_ratings.csv', help='which dataset to use')
parser.add_argument('--dim', type=int, default=64, help='dimension of entity and relation embeddings')
parser.add_argument('--l2_weight', type=float, default=1e-5, help='weight of the l2 regularization term')
parser.add_argument('--lr', type=float, default=0.001, help='learning rate')
parser.add_argument('--lr_2', type=float, default=0.0005, help='learning rate for second optimizer')
parser.add_argument('--batch_size', type=int, default=128, help='batch size')
parser.add_argument('--n_epoch', type=int, default=10, help='the number of epochs')
parser.add_argument('--hidden_layer', type=int, default=256, help='neural hidden layer')
parser.add_argument('--num_user_features', type=int, default=3, help='the number of user attributes')
parser.add_argument('--random_seed', type=int, default=2019, help='size of common item be counted')
parser.add_argument('--log_dir', type=str, default='log_dir', help='log_dir')
#random.choice(range(2000))
args = parser.parse_args()

sep = ','

if args.dataset == 'ml-1m':
    args.num_user_features = 4
    args.first_epoch = 5
    args.lr_2 = 0.0008
elif args.dataset == 'book-crossing':
    args.num_user_features = 3
    args.first_epoch = 5
elif args.dataset == 'taobao':
    args.num_user_features = 8
    args.first_epoch = 6
    args.lr_2 = 0.1
else:
    print("please specify the number of user features: num_user_features")
    


#dataset = Dataset_larger('../data/', args.dataset, args.rating_file, sep, 10)
dataset = Dataset('../data/', args.dataset, args.rating_file, sep, args)

# 1050834
data_num = dataset.data_N()
# 101285
feature_num = dataset.feature_N()
# 630500    840667
train_index, val_index = dataset.stat_info['train_test_split_index']
#print(np.concatenate((dataset[0][0:5], dataset[0][6:10])))

# split inner graphs
# Dataset(630500)
train_dataset = dataset[:train_index]
# Dataset(210167)
val_dataset = dataset[train_index:val_index]
# Dataset(210167)
test_dataset = dataset[val_index:]

# todo GPU不能用，暂时禁掉这个
# n_workers = 4
train_loader = DataLoader(train_dataset, batch_size=args.batch_size)
val_loader = DataLoader(val_dataset, batch_size=args.batch_size)
test_loader = DataLoader(test_dataset, batch_size=args.batch_size)

show_loss = False
print(f"""
datast: {args.dataset}
vector dim: {args.dim}
batch_size: {args.batch_size}
lr: {args.lr}
""")

datainfo = {}
datainfo['train'] = train_loader
datainfo['val'] = val_loader 
datainfo['test'] = test_loader
# 101285
datainfo['feature_num'] = feature_num
# [630500, 210167, 210167]
datainfo['data_num'] = [len(train_dataset), len(val_dataset), len(test_dataset)]


train(args, datainfo, show_loss)

