import torch
from torch import Tensor
import torch.nn as nn
from torch.nn import Parameter
import torch.nn.functional as F
from torch_geometric.nn.conv import MessagePassing
from torch_geometric.nn import global_mean_pool, global_add_pool, global_max_pool
from torch_geometric.utils import softmax, add_self_loops
from torch_geometric.nn.inits import glorot, zeros
from torch.nn import Linear
import numpy as np
import time
from icecream import ic


class My_GNN(MessagePassing):
    def __init__(self, dim, hidden_layer):
        super(My_GNN, self).__init__(aggr='mean')

        # construct pairwise modeling network
        # Linear(in_features=64, out_features=256, bias=True)
        self.lin1 = nn.Linear(dim * 2, hidden_layer)
        # Linear(in_features=256, out_features=64, bias=True)
        self.lin2 = nn.Linear(hidden_layer, dim)
        self.act = nn.ReLU()
        self.drop = nn.Dropout(p=0.5)

        self.lin3 = nn.Linear(dim, hidden_layer)
        self.lin4 = nn.Linear(dim, hidden_layer)

    # x: torch.Size([1374, 1, 64])    edge_index: torch.Size([2, 9556])  edge_weight: None
    def forward(self, x, edge_index, edge_weight=None):
        # x has shape [N, dim]
        # edge_index has shape [2, E]
        # try:
        #    return self.propagate(edge_index, x=x, edge_weight=edge_weight)
        # except:
        x = x.squeeze()
        return self.propagate(edge_index, x=x, edge_weight=edge_weight)

    # x_i和x_j应该是每条边涉及的两个节点的表示  9556这个数字是边的数量其2倍, 查看右边链接的"MessagePassing.propagate(edge_index, size=None, **kwargs)"处的解读，
    # 可知边是(N,N)的邻接矩阵，见 https://pytorch-geometric.readthedocs.io/en/latest/notes/create_gnn.html
    # torch.Size([9556, 64])    torch.Size([9556, 64])  None
    def message(self, x_i, x_j, edge_weight):
        # x_i has shape [E, dim]
        # x_j has shape [E, dim]

        # pairwise analysis
        # torch.Size([9556, 64])
        # pairwise_analysis = x_i * x_j
        pairwise_analysis = self.lin1(torch.cat((x_i, x_j), 1)) + self.lin3(x_i + x_j) + self.lin4(x_i * x_j)
        pairwise_analysis = self.act(pairwise_analysis)
        # torch.Size([9556, 64])
        pairwise_analysis = self.lin2(pairwise_analysis)
        # pairwise_analysis = self.drop(pairwise_analysis)
        # pairwise_analysis = x_i * x_j

        if edge_weight != None:
            interaction_analysis = pairwise_analysis * edge_weight.view(-1, 1)
        else:
            interaction_analysis = pairwise_analysis

        return interaction_analysis

    def update(self, aggr_out):
        # aggr_out has shape [N, dim]
        return aggr_out


# class inner_GNN(MessagePassing):
#     def __init__(self, dim, hidden_layer):
#         super(inner_GNN, self).__init__(aggr='mean')
#
#         # construct pairwise modeling network
#         # Linear(in_features=64, out_features=256, bias=True)
#         self.lin1 = nn.Linear(dim, hidden_layer)
#         # Linear(in_features=256, out_features=64, bias=True)
#         self.lin2 = nn.Linear(hidden_layer, dim)
#         self.act = nn.ReLU()
#         self.drop = nn.Dropout(p=0.5)
#
#     # x: torch.Size([1374, 1, 64])    edge_index: torch.Size([2, 9556])  edge_weight: None
#     def forward(self, x, edge_index, edge_weight=None):
#         # x has shape [N, dim]
#         # edge_index has shape [2, E]
#         # try:
#         #    return self.propagate(edge_index, x=x, edge_weight=edge_weight)
#         # except:
#         x = x.squeeze()
#         return self.propagate(edge_index, x=x, edge_weight=edge_weight)
#
#     # x_i和x_j应该是每条边涉及的两个节点的表示  9556这个数字是边的数量其2倍, 查看右边链接的"MessagePassing.propagate(edge_index, size=None, **kwargs)"处的解读，
#     # 可知边是(N,N)的邻接矩阵，见 https://pytorch-geometric.readthedocs.io/en/latest/notes/create_gnn.html
#     # torch.Size([9556, 64])    torch.Size([9556, 64])  None
#     def message(self, x_i, x_j, edge_weight):
#         # x_i has shape [E, dim]
#         # x_j has shape [E, dim]
#
#         # pairwise analysis
#         # torch.Size([9556, 64])
#         pairwise_analysis = x_i * x_j
#         pairwise_analysis = self.lin1(pairwise_analysis)
#         pairwise_analysis = self.act(pairwise_analysis)
#         # torch.Size([9556, 64])
#         pairwise_analysis = self.lin2(pairwise_analysis)
#         pairwise_analysis = self.drop(pairwise_analysis)
#         # pairwise_analysis = x_i * x_j
#
#         if edge_weight != None:
#             interaction_analysis = pairwise_analysis * edge_weight.view(-1, 1)
#         else:
#             interaction_analysis = pairwise_analysis
#
#         return interaction_analysis
#
#     def update(self, aggr_out):
#         # aggr_out has shape [N, dim]
#         return aggr_out


class cross_GNN(MessagePassing):
    def __init__(self, dim, hidden_layer):
        super(cross_GNN, self).__init__(aggr='mean')

    # torch.Size([1374, 1, 64])  torch.Size([2, 5940])  None
    def forward(self, x, edge_index, edge_weight=None):
        # x has shape [N, dim]
        # edge_index has shape [2, E]
        # try:
        #    return self.propagate(edge_index, x=x, edge_weight=edge_weight)
        # except:
        x = x.squeeze()
        return self.propagate(edge_index, x=x, edge_weight=edge_weight)

    # def message(self, edge_index_i, x_i, x_j, edge_weight, num_nodes):
    # torch.Size([5940, 64])    torch.Size([5940, 64])  None
    def message(self, x_i, x_j, edge_weight):
        # x_i has shape [E, dim]
        # x_j has shape [E, dim]

        # pairwise analysis
        # pairwise_analysis = x_i * x_j
        pairwise_analysis = torch.sigmoid(torch.sum(x_i * x_j, 1).unsqueeze(1)) * x_j

        if edge_weight != None:
            interaction_analysis = pairwise_analysis * edge_weight.view(-1, 1)
        else:
            interaction_analysis = pairwise_analysis

        return interaction_analysis

    def update(self, aggr_out):
        # aggr_out has shape [N, dim]
        return aggr_out


class GMCF(nn.Module):
    """
    GMCF main model
    """

    def __init__(self, args, n_features, device):
        super(GMCF, self).__init__()

        # 101285  这个是user个数加item个数加User attr个数加Item attr个数
        self.n_features = n_features
        # 64
        self.dim = args.dim
        # 256
        self.hidden_layer = args.hidden_layer
        self.device = device
        # 128
        self.batch_size = args.batch_size
        # 3
        self.num_user_features = args.num_user_features

        # Embedding(101286, 64)
        self.feature_embedding = nn.Embedding(self.n_features + 1, self.dim)
        # self.feature_embedding.weight.data.normal_(0.0,0.01)
        # self.inner_gnn = inner_GNN(self.dim, self.hidden_layer)
        # self.outer_gnn = cross_GNN(self.dim, self.hidden_layer)
        self.my_gnn = My_GNN(self.dim, self.hidden_layer)

        # Embedding(101286, 1)
        self.node_weight = nn.Embedding(self.n_features + 1, 1)
        self.node_weight.weight.data.normal_(0.0, 0.01)

        # GRU(64, 64, dropout=0.5)
        self.update_f = nn.GRU(input_size=self.dim, hidden_size=self.dim, dropout=0.5)
        # Linear(in_features=64, out_features=1, bias=False)
        self.g = nn.Linear(self.dim, 1, bias=False)
        self.att_1 = nn.Linear(2, 2)

    def forward(self, data, is_training=True):
        # does not conduct link prediction, use all interactions

        # torch.Size([1374, 1])  平均一个样本有10个节点
        node_id = data.x.to(self.device)
        # Tensor(1374) 1374个里面每个元素属于哪个batch，可调试时观察一下值就知道了
        batch = data.batch

        # handle pointwise features
        # Tensor(1374)
        node_w = torch.squeeze(self.node_weight(node_id))
        # Tensor(128)  看官网就是每个batch内图节点的累加
        sum_weight = global_add_pool(node_w, batch)

        # torch.Size([1374, 1, 64])
        node_emb = self.feature_embedding(node_id)
        # torch.Size([2, 9556])   值在0到1373之间
        inner_edge_index = data.edge_index
        # torch.Size([2, 5940])  每个样本图的cross interactions边
        outer_edge_index = torch.transpose(data.edge_attr, 0, 1)
        # torch.Size([2, 5940])  值的范围看起来是0到8，这是因为最后一个样本图的节点数是9，outer_edge_index中边的索引是图内部的编号，这大概是为什么需要offset了
        outer_edge_index = self.outer_offset(batch, self.num_user_features, outer_edge_index)

        # torch.Size([1374, 64])  参考4.3.1节的式2及其周边
        # inner_node_message = self.inner_gnn(node_emb, inner_edge_index)
        # torch.Size([1374, 64])  参考4.3.2节的式3及其周边
        # outer_node_message = self.outer_gnn(node_emb, outer_edge_index)
        my_node_message = self.my_gnn(node_emb, outer_edge_index)

        # Tensor(1374)
        new_batch = self.split_batch(batch, self.num_user_features)
        # torch.Size([256, 64])
        updated_graph_1 = torch.squeeze(global_mean_pool(node_emb, new_batch)) + torch.squeeze(global_max_pool(node_emb, new_batch))
        # updated_graph_2 = torch.squeeze(global_mean_pool(outer_node_message, new_batch)) + torch.squeeze(global_max_pool(outer_node_message, new_batch))
        updated_graph_3 = torch.squeeze(global_mean_pool(my_node_message, new_batch)) + torch.squeeze(global_max_pool(my_node_message, new_batch))
        # torch.Size([256, 3, 64])
        updated_node_input = torch.cat((updated_graph_1.unsqueeze(1), updated_graph_3.unsqueeze(1)), 1)
        # torch.Size([256, 64])
        updated_graph = (torch.sigmoid(self.att_1(updated_node_input.mean(-1))).unsqueeze(2) * updated_node_input).mean(-2)
        # torch.Size([256, 64])
        # updated_graph = torch.squeeze(global_mean_pool(updated_node, new_batch))
        # torch.Size([128, 64])  torch.Size([128, 64])  torch.split看官网很简单，就是分成几个chunk
        item_graphs, user_graphs = torch.split(updated_graph, int(updated_graph.size(0) / 2))

        # torch.Size([128, 1])  式5下面的y', 但又不完全一样
        y = torch.unsqueeze(torch.sum(user_graphs * item_graphs, 1) + sum_weight, 1)
        y = torch.sigmoid(y)

        return y

    # user_node_num: 3
    def split_batch(self, batch, user_node_num):
        """
        split batch id into user nodes and item nodes 
        """
        # Tensor(1374)
        ones = torch.ones_like(batch)
        # Tensor(128)  [13,13,9,16,...,8,9]
        nodes_per_graph = global_add_pool(ones, batch)
        # Tensor(128)  [0,13,26,35,51,...,1357,1365]
        cum_num = torch.cat((torch.LongTensor([0]).to(self.device), torch.cumsum(nodes_per_graph, dim=0)[:-1]))
        # list(3)   cum_num_list[0]: Tensor(128)
        # [[0,13,26,...,1357,1365],[1,14,27,...,1358,1366],[2,15,28,...]]
        # 这个的作用是标记user属性的位置，因为此数据集user有3个属性，所以有3行，第一行标记第一个属性的位置，第二行标记第二个属性的位置，一次类推
        cum_num_list = [cum_num + i for i in range(user_node_num)]
        # Tensor(384)   [0,13,26,...,]
        multi_hot = torch.cat(cum_num_list)
        # temp_1 = F.one_hot(multi_hot, ones.size(0))  # torch.Size([384, 1374])
        # Tensor(1374)   下面torch.sum后所有为1的位置就是user属性的位置，然后都乘以128
        test = torch.sum(F.one_hot(multi_hot, ones.size(0)), dim=0) * (torch.max(batch) + 1)
        # 通过final[:50]打印前50个值发现，属于user的属性都是三个，且值大于等于128，小于128的都是item的属性
        # tensor([128, 128, 128,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, 129,
        #         129, 129,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1, 130, 130,
        #         130,   2,   2,   2,   2,   2,   2, 131, 131, 131,   3,   3,   3,   3,
        #           3,   3,   3,   3,   3,   3,   3,   3])
        # 观察上面值分布，第一个样本包含13个属性节点，前3个都是user属性，剩下都是标记为item属性的0，第二个样本也是13个属性，前3个位user属性，第三个样本9个属性，前三个user属性，后6个item属性
        final = batch + test

        return final

    # batch: Tensor(1374)   user_node_num: 3    outer_edge_index: torch.Size([2, 5940])  值的范围看起来是0到8，这是因为最后一个样本图的节点数是9
    def outer_offset(self, batch, user_node_num, outer_edge_index):
        ones = torch.ones_like(batch)
        # Tensor(128)  [13,13,9,16,11,...8,9]
        nodes_per_graph = global_add_pool(ones, batch)
        # Tensor(128)  [60,60,36,78,48,...,30,36]  这个计算的应该是每个图中user属性和item属性之间的连边数乘以2，
        # nodes_per_graph - user_node_num应该是item属性的个数，user_node_num是user属性的个数
        inter_per_graph = (nodes_per_graph - user_node_num) * user_node_num * 2
        # Tensor(128)  [0,13,26,35,51,...,1357,1365]
        cum_num = torch.cat((torch.LongTensor([0]).to(self.device), torch.cumsum(nodes_per_graph, dim=0)[:-1]))
        # Tensor(5940)
        # temp = torch.repeat_interleave(cum_num, inter_per_graph, dim=0)
        # temp_1 = torch.tensor([1, 2, 3])
        # temp_2 = torch.repeat_interleave(temp_1, torch.tensor([2, 3, 4]), dim=0)
        # torch.Size([2, 5940])   第一行部分值为: [0,0,0,...,1365,1365,1365]   第一行和第二行因为repeat相同
        offset_list = torch.repeat_interleave(cum_num, inter_per_graph, dim=0).repeat(2, 1)
        # torch.Size([2, 5940])  值的范围在0到1373
        outer_edge_index_offset = outer_edge_index + offset_list
        return outer_edge_index_offset
